@@ -191,6 +191,8 @@ v3_t triangulate_n(int num_points,
     int i;
     double error;
 
+    v3_t r;
+
     for (i = 0; i < num_points; i++) {
 	int Roff = 9 * i;
 	int row = 6 * i;
@@ -266,7 +268,7 @@ v3_t triangulate_n(int num_points,
 	*error_out = error;
     }
 
-    v3_t r = v3_new(x[0], x[1], x[2]);
+    r = v3_new(x[0], x[1], x[2]);
 
     free(A);
     free(b);
@@ -567,8 +569,8 @@ int find_extrinsics_essential_multipt(double *E, int n,
             c2_neg++;
     }
 
-    printf("[1] c1_pos: %d, c1_neg: %d\n", c1_pos, c1_neg);
-    printf("[1] c2_pos: %d, c2_neg: %d\n", c2_pos, c2_neg);
+    // printf("[1] c1_pos: %d, c1_neg: %d\n", c1_pos, c1_neg);
+    // printf("[1] c2_pos: %d, c2_neg: %d\n", c2_pos, c2_neg);
 
     // EDIT!!!
     if (c1_pos < c1_neg && c2_pos < c2_neg) {
@@ -606,8 +608,8 @@ int find_extrinsics_essential_multipt(double *E, int n,
                 c2_neg++;
         }
 
-        printf("[2] c1_pos: %d, c1_neg: %d\n", c1_pos, c1_neg);
-        printf("[2] c2_pos: %d, c2_neg: %d\n", c2_pos, c2_neg);
+        // printf("[2] c1_pos: %d, c1_neg: %d\n", c1_pos, c1_neg);
+        // printf("[2] c2_pos: %d, c2_neg: %d\n", c2_pos, c2_neg);
         
         if (c1_pos < c1_neg && c2_pos < c2_neg) { // EDIT!!!
             memcpy(R, Rb, 9 * sizeof(double));
@@ -621,12 +623,12 @@ int find_extrinsics_essential_multipt(double *E, int n,
 
             // Q = triangulate(p1, p2, I, t0, R, t, &error);
         } else {
-            printf("[find_extrinsics] Error: no case found!\n");
+            fprintf(stderr, "[find_extrinsics] Error: no case found!\n");
             return 0;
         }
     }
 
-    printf("[find_extrinsics] error: %0.3e\n", error);
+    // printf("[find_extrinsics] error: %0.3e\n", error);
 
     // matrix_product(3, 3, 3, 1, R, Q.p, tmp2);
     // matrix_sum(3, 1, 3, 1, tmp2, t, tmp2);
@@ -1057,6 +1059,15 @@ int find_projection_3x4_ransac(int num_pts, v3_t *points, v2_t *projs,
 	       max_inliers, num_pts);
 	printf("[find_projection_3x4_ransac] error = %0.3f\n", 
 	       sqrt(max_error / max_inliers));
+
+        if (max_inliers < 6) {
+            printf("[find_projection_3x4_ransac] "
+                   "Too few inliers to continue.\n");
+            
+            free(inliers);
+
+            return -1;
+        }
 	
 	/* Do the final least squares minimization */
 
