@@ -23,6 +23,7 @@
 #include <stdlib.h>
 #include <stdio.h>
 #include <string>
+#include <time.h>
 
 #include "BundlerApp.h"
 #include "Bundle.h"
@@ -44,7 +45,7 @@
 #include "util.h"
 #include "vector.h"
 
-#ifndef WIN32_MVC
+#ifndef WIN32
 #include <ext/hash_map>
 #else
 #include <hash_map>
@@ -271,7 +272,7 @@ void TwoFrameModel::WriteWithProjections(FILE *f,
 #endif
 }
 
-
+#if 0
 void WriteModelsProjections(ModelMap &models, int num_images, 
                             const std::vector<TrackData> &track_data,
                             std::vector<ImageData> &image_data, char *out_file)
@@ -294,6 +295,7 @@ void WriteModelsProjections(ModelMap &models, int num_images,
     } else {
         fprintf(f, "%d\n", num_images);
 
+        // FIXME LOOP
         for (int i = 0; i < num_images; i++) {
             ImageData &img_i = image_data[i];
             
@@ -301,14 +303,14 @@ void WriteModelsProjections(ModelMap &models, int num_images,
                 image_data[i].LoadKeys(false);
 
             for (int j = i+1; j < num_images; j++) {
-                if (models.Exists(i * num_images + j)) {
+                if (models.Contains(GetMatchIndex(i, j))) {
                     if (!preload_keys)
                         image_data[j].LoadKeys(false);
 
                     ImageData &img_j = image_data[j];
 
                     fprintf(f, "%d %d\n", i, j);
-                    models.Get(i * num_images + j).
+                    models.GetModel(GetMatchIndex(i,j)).
                         WriteWithProjections(f, track_data, i, j, 
                                              img_i, img_j);
 
@@ -330,6 +332,7 @@ void WriteModelsProjections(ModelMap &models, int num_images,
         }
     }
 }
+#endif
 
 #if 0
 static void FixPEdges(int num_images, ModelMap &models, PEdgeMap &p_edges)
@@ -361,11 +364,12 @@ static void FixScaffoldEdges(int num_images, ModelMap &models)
 {
     const double MIN_ANGLE = 1.5;
     
+    // FIXME LOOP ?
     for (int i = 0; i < num_images; i++) {
         for (int j = i+1; j < num_images; j++) {
-            int idx = i * num_images + j;
-            if (models.Exists(idx)) {
-                TwoFrameModel &m = models.Get(idx);
+            MatchIndex idx = GetMatchIndex(i,j); // i * num_images + j;
+            if (models.Contains(idx)) {
+                TwoFrameModel &m = models.GetModel(idx);
 
                 if (m.m_angle < MIN_ANGLE) {
                     printf("[FixScaffoldEdges] "
@@ -382,7 +386,7 @@ static void FixScaffoldEdges(int num_images, ModelMap &models)
                         models.erase(idx);
                     }
 #else
-                    models.Remove(idx);
+                    models.RemoveModel(idx);
 #endif
                 }
             }
@@ -390,11 +394,11 @@ static void FixScaffoldEdges(int num_images, ModelMap &models)
     }
 }
 
-ModelMap SkeletalApp::BundleAllPairs(char *out_file, PEdgeMap &p_edges_out,
+ModelMap SkeletalApp::BundleAllPairs(char *out_file, 
                                      bool bundle_from_tracks, 
                                      bool detect_duplicates) 
 {
-    int num_images = GetNumImages();
+    unsigned int num_images = GetNumImages();
 
     if (out_file != NULL) {
         FILE *f = fopen(out_file, "r");
@@ -403,45 +407,44 @@ ModelMap SkeletalApp::BundleAllPairs(char *out_file, PEdgeMap &p_edges_out,
         if (strstr(out_file, "pruned"))
             load_pruned = true;
 
-        FILE *fp;
-        if (!load_pruned)
-            fp = fopen("pedges.out", "r");
-        else
-            fp = fopen("pedges.pruned.out", "r");
-
-        if (f != NULL && fp != NULL) {
+        if (f != NULL) {
             ModelMap models = ReadModels(f);
-            p_edges_out = ReadPEdges(fp, num_images);
+            // p_edges_out = ReadPEdges(fp, num_images);
             FixScaffoldEdges(num_images, models);
+            ThresholdTwists(num_images, models, m_image_data, true);
             fclose(f);
-            fclose(fp);
             return models;
         }
     }
 
     /* Sort images by connectivity */
     double *connectivity = new double[num_images];
-    bool *connected = new bool[num_images * num_images];
+    // bool *connected = new bool[num_images * num_images];
 
 #define MATCH_THRESHOLD 28   // 16
 
-    for (int i = 0; i < num_images; i++)
+    for (unsigned int i = 0; i < num_images; i++)
         connectivity[i] = 0.0;
 
-    for (int i = 0; i < num_images * num_images; i++) {
-        connected[i] = false;
-    }
+    // for (int i = 0; i < num_images * num_images; i++) {
+    //     connected[i] = false;
+    // }
 
-    int num_connections = 0;
-    for (int i = 0; i < num_images; i++) {
+    unsigned long long num_connections = 0;
+    for (unsigned int i = 0; i < num_images; i++) {
         if (!m_image_data[i].m_has_init_focal)
             continue;
 
-        for (int j = i+1; j < num_images; j++) {
+        MatchAdjList::iterator iter;
+
+        // for (int j = i+1; j < num_images; j++) {
+        for (iter = m_matches.Begin(i); iter != m_matches.End(i); iter++) {
+            unsigned int j = iter->m_index;
+
             if (!m_image_data[j].m_has_init_focal)
                 continue;
 
-            if (i == j) 
+            if (i >= j) 
                 continue;
 
             int num_matches; 
@@ -457,8 +460,8 @@ ModelMap SkeletalApp::BundleAllPairs(char *out_file, PEdgeMap &p_edges_out,
             if (num_matches >= MATCH_THRESHOLD) {
                 connectivity[i] += 1.0;
                 connectivity[j] += 1.0;
-                connected[i * num_images + j] = true;
-                connected[j * num_images + i] = true;
+                // connected[i * num_images + j] = true;
+                // connected[j * num_images + i] = true;
                 num_connections++;
             }
         }
@@ -475,24 +478,29 @@ ModelMap SkeletalApp::BundleAllPairs(char *out_file, PEdgeMap &p_edges_out,
 
     delete [] connectivity;
 #else
-    for (int i = 0; i < num_images; i++) {
+    for (unsigned int i = 0; i < num_images; i++) {
         printf("connect[ %d ] = %0.3f\n", i, connectivity[i]);
     }
 #endif
 
     double *match_array = new double[num_connections];
-    int *indices = new int[num_connections];
+    unsigned long *indices = new unsigned long [num_connections];
     
     int count = 0;
-    for (int i = 0; i < num_images; i++) {
+    for (unsigned int i = 0; i < num_images; i++) {
         if (!m_image_data[i].m_has_init_focal)
             continue;
 
-        for (int j = i+1; j < num_images; j++) {
+        // for (int j = i+1; j < num_images; j++) {
+        MatchAdjList::iterator iter;
+
+        for (iter = m_matches.Begin(i); iter != m_matches.End(i); iter++) {
+            unsigned int j = iter->m_index;
+
             if (!m_image_data[j].m_has_init_focal)
                 continue;
 
-            if (i == j) 
+            if (i >= j) 
                 continue;
 
             int num_matches;
@@ -507,14 +515,16 @@ ModelMap SkeletalApp::BundleAllPairs(char *out_file, PEdgeMap &p_edges_out,
             
             if (num_matches >= MATCH_THRESHOLD) {
                 match_array[count] = (double) num_matches;
-                indices[count] = i * num_images + j;
+                indices[count] = 
+                    (unsigned long long) i * num_images + 
+                    (unsigned long long) j;
                 count++;
             }
         }
     }
     
     int *order = new int[num_connections];
-    qsort_perm(num_connections, match_array, order);
+    qsort_perm((int) num_connections, match_array, order);
     delete [] match_array;
 
     bool preload_keys = true;
@@ -522,7 +532,7 @@ ModelMap SkeletalApp::BundleAllPairs(char *out_file, PEdgeMap &p_edges_out,
         preload_keys = false;
 
     if (preload_keys) {
-        for (int i = 0; i < num_images; i++) {
+        for (unsigned int i = 0; i < num_images; i++) {
             if (m_image_data[i].m_has_init_focal) {
                 printf("[BundleAllPairs] Loading keys for image %d\n", i);
                 m_image_data[i].LoadKeys(false, !m_optimize_for_fisheye);
@@ -539,20 +549,19 @@ ModelMap SkeletalApp::BundleAllPairs(char *out_file, PEdgeMap &p_edges_out,
     }
 
     ModelMap models(num_images);
-    PEdgeMap p_edges;
 
     int *depends = new int[num_images];
 
-    for (int i = 0; i < num_images; i++) {
+    for (unsigned int i = 0; i < num_images; i++) {
         depends[i] = -1;
     }
     
     // for (int i = 0; i < num_images; i++) {
-    for (int i = 0; i < num_connections; i++) {
-        int index = indices[order[i]];
+    for (unsigned long i = 0; i < num_connections; i++) {
+        unsigned long long index = indices[order[i]];
 
-        int i1 = index / num_images;
-        int i2 = index - i1 * num_images;
+        unsigned long i1 = index / num_images;
+        unsigned long i2 = index - i1 * num_images;
         
         if (connectivity[i2] > connectivity[i1]) {
             int tmp = i1;
@@ -589,7 +598,7 @@ ModelMap SkeletalApp::BundleAllPairs(char *out_file, PEdgeMap &p_edges_out,
                 continue;
         }
 
-        printf("[SifterApp::BundleAllPairs] Bundling (%d,%d)\n", i1, i2);
+        printf("[SifterApp::BundleAllPairs] Bundling (%lu,%lu)\n", i1, i2);
         fflush(stdout);
 
         if (!preload_keys) {
@@ -597,18 +606,17 @@ ModelMap SkeletalApp::BundleAllPairs(char *out_file, PEdgeMap &p_edges_out,
             SetTracks(i2);
         }
             
-        int i_min = MIN(i1, i2);
-        int i_max = MAX(i1, i2);
+        unsigned int i_min = MIN(i1, i2);
+        unsigned int i_max = MAX(i1, i2);
 
         TwoFrameModel model;
-        bool p_edge = false;
 
         clock_t start = clock();
         double angle = 0.0;
         int num_matches = 0;
         bool success = 
             BundleTwoFrame(i_min, i_max, 
-                           &model, angle, num_matches, p_edge, 
+                           &model, angle, num_matches, 
                            bundle_from_tracks);
         fflush(stdout);
 
@@ -624,12 +632,29 @@ ModelMap SkeletalApp::BundleAllPairs(char *out_file, PEdgeMap &p_edges_out,
 
             // bool found = false;
             int num_found = 0;
-            for (int k = 0; k < num_images; k++) {
-                if (k == i1 || k == i2)
-                    continue;                
+            // for (int k = 0; k < num_images; k++) {
+            //     if (k == i1 || k == i2)
+            //         continue;
 
-                if (!connected[i1 * num_images + k] &&
-                    connected[i2 * num_images + k]) {
+            MatchAdjList::iterator iter;
+            
+            for (iter = m_matches.Begin(i2); 
+                 iter != m_matches.End(i2); 
+                 iter++) {
+
+                unsigned long k = iter->m_index;
+
+                if (k == i1 || k == i2)
+                    continue;
+
+                // if (!connected[i1 * num_images + k] &&
+                //     connected[i2 * num_images + k]) {
+                if ((bundle_from_tracks && 
+                     GetNumTrackMatches(i1,k) < MATCH_THRESHOLD &&
+                     GetNumTrackMatches(i2,k) >= MATCH_THRESHOLD) ||
+                    (!bundle_from_tracks &&
+                     (GetNumMatches(i1,k) < MATCH_THRESHOLD || 
+                      GetNumMatches(i2,k) >= MATCH_THRESHOLD))) {
 
                     int num_matches_i2k;
                     if (bundle_from_tracks)
@@ -638,8 +663,8 @@ ModelMap SkeletalApp::BundleAllPairs(char *out_file, PEdgeMap &p_edges_out,
                         num_matches_i2k = GetNumMatches(i2, k);
 
                     if (num_matches_i2k > 64) {
-                        printf("  Image %d is connected to %d (%d matches) "
-                               "but not %d\n",
+                        printf("  Image %lu is connected to %lu (%d matches) "
+                               "but not %lu\n",
                                k, i2, num_matches_i2k, i1);
 
                         // found = true;
@@ -653,33 +678,44 @@ ModelMap SkeletalApp::BundleAllPairs(char *out_file, PEdgeMap &p_edges_out,
             if (num_found < 1 /*MAX(1.0, 0.002 * num_images)*/) {
                 /* Make node i2 dependent on node i1 */
                 printf("[SifterApp::BundleAllPairs] "
-                       "Node %d depends on node %d\n", i2, i1);
+                       "Node %lu depends on node %lu\n", i2, i1);
 
                 dependent = true;
                 depends[i2] = i1;
 
                 /* Get rid of other models with i2 */
-                for (int k = 0; k < i2; k++) {
-                    int idx = k * num_images + i2;
-                    if (models.Exists(idx)) {
-                        printf("  Removed model (%d,%d)\n", k, i2);
-                        models.Remove(idx);
+                // for (int k = 0; k < i2; k++) {
+                for (iter = m_matches.Begin(i2); 
+                     iter != m_matches.End(i2); 
+                     iter++) {
+                    
+                    unsigned long k = iter->m_index;
+
+                    // int idx = k * num_images + i2;
+                    MatchIndex idx = GetMatchIndexUnordered(k, i2);
+                    if (models.Contains(idx)) {
+                        printf("  Removed model (%lu,%lu)\n", k, i2);
+                        models.RemoveModel(idx);
                     }
                 }
                 
+#if 0
+                // FIXME LOOP
                 for (int k = i2 + 1; k < num_images; k++) {
-                    int idx = i2 * num_images + k;
-                    if (models.Exists(idx)) {
+                    // int idx = i2 * num_images + k;
+                    MatchIndex idx = GetMatchIndex(i2, k);
+                    if (models.Contains(idx)) {
                         printf("  Removed model (%d,%d)\n", i2, k);
-                        models.Remove(idx);
-                    }         
+                        models.RemoveModel(idx);
+                    }
                 }
+#endif
             }
         }
 
         if (!dependent && success) {
             printf("[SifterApp::BundleAllPairs] "
-                   "(%d,%d) successfully bundled\n", i1, i2);
+                   "(%lu,%lu) successfully bundled\n", i1, i2);
             
             if (model.ComputeTrace(true) < 0.0) {
                 printf("  Error! trace2 < 0.0!\n");
@@ -688,15 +724,11 @@ ModelMap SkeletalApp::BundleAllPairs(char *out_file, PEdgeMap &p_edges_out,
             if (model.ComputeTrace(false) < 0.0) {
                 printf("  Error! trace1 < 0.0!\n");
             } else {
-                models.Add(i_min * num_images + i_max, model);
+                MatchIndex idx = GetMatchIndex(i_min, i_max);
+                models.AddModel(idx, model);
             }
         } else if (!dependent) {
-            if (p_edge) {
-                printf("[BundleAllPairs] (%d,%d) is a p-edge\n", i1, i2);
-                p_edges[i_min * num_images + i_max] = true;
-            } else {
-                printf("[BundleAllPairs] (%d,%d) bundle FAILED\n", i1, i2);
-            }
+            printf("[BundleAllPairs] (%lu,%lu) bundle FAILED\n", i1, i2);
         }
 
         clock_t end = clock();
@@ -719,13 +751,14 @@ ModelMap SkeletalApp::BundleAllPairs(char *out_file, PEdgeMap &p_edges_out,
 
     if (out_file != NULL) {
         WriteModels(models, num_images, out_file);
-        WriteModelsProjections(models, num_images, m_track_data, m_image_data,
-                               "pairs.proj.out");
-        WritePEdges(p_edges, num_images, "pedges.out");
+        // WriteModelsProjections(models, num_images, 
+        //                        m_track_data, m_image_data, 
+        //                        "pairs.proj.out");
+        // WritePEdges(p_edges, num_images, "pedges.out");
     }
     fflush(stdout);
 
-    p_edges_out = p_edges;
+    // p_edges_out = p_edges;
 
     delete [] depends;
 #if 0
@@ -832,7 +865,6 @@ double BundlerApp::RunSFMNecker(int i1, int i2,
 
 bool BundlerApp::BundleTwoFrame(int i1, int i2, TwoFrameModel *model,
                                 double &angle_out, int &num_pts_out, 
-                                bool &p_edge, 
                                 bool bundle_from_tracks) 
 {
     const double TERM_THRESH = 1.0e-12;
@@ -840,8 +872,6 @@ bool BundlerApp::BundleTwoFrame(int i1, int i2, TwoFrameModel *model,
     // assert(!m_estimate_distortion && !m_fixed_focal_length);
     assert(!m_fixed_focal_length);
 
-    p_edge = false;
-
     if (!m_image_data[i1].m_has_init_focal || 
         !m_image_data[i2].m_has_init_focal) {
 
@@ -2089,10 +2119,123 @@ bool BundlerApp::BundleTwoFrame(int i1, int i2, TwoFrameModel *model,
     return true;
 }
 
-void SkeletalApp::ComputeCameraCovariance()
+void BundlerApp::ComputeCameraCovariance()
+{
+    int num_images = GetNumImages();
+
+    int cnp = (m_estimate_distortion) ? 9 : 7;
+    int num_vars = cnp * num_images;
+    double *S = new double[num_vars * num_vars];
+
+    /* Add constraints */
+    m_use_point_constraints = true;
+    m_point_constraint_weight = 1000.0;
+
+    int num_points = (int) m_point_data.size();
+    m_point_constraints = new v3_t[num_points];
+    
+    for (int i = 0; i < num_points; i++) {
+        PointData &p = m_point_data[i];
+        m_point_constraints[i] = v3_new(p.m_pos[0], p.m_pos[1], p.m_pos[2]);
+    }
+
+    ReRunSFM(S);
+
+#if 0
+    /* Debugging stuff */
+    for (int i = 1; i < num_images + 1; i++) {
+        int num_vars_sub = i * cnp;
+        double *Ssub = new double[num_vars_sub * num_vars_sub];
+
+        for (int j = 0; j < num_vars_sub; j++) {
+            memcpy(Ssub + j * num_vars_sub, S + j * num_vars, 
+                   sizeof(double) * num_vars_sub);
+        }
+
+        double *U, *S, *VT;
+        U = new double[num_vars_sub * num_vars_sub];
+        S = new double[num_vars_sub];
+        VT = new double[num_vars_sub * num_vars_sub];
+        
+        dgesvd_driver(num_vars_sub, num_vars_sub, Ssub, U, S, VT);
+        printf("Sing. values [round %d]\n", i);
+        for (int j = 0; j < num_vars_sub; j++) {
+            printf("   S[%d] = %0.5e\n", j, S[j]);
+        }
+
+        printf("Smallest eigenvector [round %d]:\n", i);
+        matrix_print(i, cnp, VT + (num_vars_sub-1) * num_vars_sub);
+
+        delete [] U;
+        delete [] S;
+        delete [] VT;
+
+        double *Sinv_sub = new double[num_vars_sub * num_vars_sub];
+        matrix_invert(num_vars_sub, Ssub, Sinv_sub);
+
+        for (int j = 0; j < i; j++) {
+            int row0 = (cnp * j + 0) * num_vars_sub;
+            int row1 = (cnp * j + 1) * num_vars_sub;
+            int row2 = (cnp * j + 2) * num_vars_sub;
+            int off = cnp * j;
+
+            double C[9] = { Sinv_sub[row0+off+0], Sinv_sub[row0+off+1], Sinv_sub[row0+off+2],
+                            Sinv_sub[row1+off+0], Sinv_sub[row1+off+1], Sinv_sub[row1+off+2],
+                            Sinv_sub[row2+off+0], Sinv_sub[row2+off+1], Sinv_sub[row2+off+2] };
+            
+            printf("Covariance[%d] [Image %d]:\n", i, j);
+            matrix_print(3, 3, C);
+        }
+
+        delete [] Ssub;
+        delete [] Sinv_sub;
+    }
+#endif
+
+    /* Invert S */
+    double *Sinv = new double[num_vars * num_vars];
+    matrix_invert(num_vars, S, Sinv);
+    
+    FILE *f = fopen("covariance.txt", "w");
+    if (f == NULL) {
+        printf("[ComputeCameraCovariance] Error opening file %s for writing\n",
+               "covariance.txt");
+        return;
+    }
+
+    int count = 0;
+    for (int i = 0; i < num_images; i++) {
+        if (m_image_data[i].m_camera.m_adjusted) {
+            int row0 = (cnp * count + 0) * num_vars;
+            int row1 = (cnp * count + 1) * num_vars;
+            int row2 = (cnp * count + 2) * num_vars;
+            int off = cnp * count;
+
+            double C[9] = 
+                { Sinv[row0+off+0], Sinv[row0+off+1], Sinv[row0+off+2],
+                  Sinv[row1+off+0], Sinv[row1+off+1], Sinv[row1+off+2],
+                  Sinv[row2+off+0], Sinv[row2+off+1], Sinv[row2+off+2] };
+
+            fprintf(f, "%d\n", i);
+            fprintf(f, "%0.6e %0.6e %0.6e "
+                    "%0.6e %0.6e %0.6e "
+                    "%0.6e %0.6e %0.6e\n", 
+                    C[0], C[1], C[2], C[3], C[4], C[5], C[6], C[7], C[8]);
+            fprintf(f, "%0.6e\n", C[0] + C[4] + C[8]);
+
+            printf("Covariance [Image %d]:\n", i);
+            matrix_print(3, 3, C);
+
+            count++;
+        }
+    }
+}
+
+#if 0
+void BundlerApp::ComputeCameraCovariance()
 {
     if (m_covariance_fix1 == -1 || m_covariance_fix2 == -1) {
-        printf("[SifterApp::ComputeCovariance] "
+        printf("[SifterApp::ComputeCameraCovariance] "
                "Two images need to be specified.\n");
     }
     
@@ -2216,6 +2359,13 @@ void SkeletalApp::ComputeCameraCovariance()
     double *Sinv = new double[num_vars * num_vars];
     matrix_invert(num_vars, S, Sinv);
     
+    FILE *f = fopen("covariance.txt", "w");
+    if (f == NULL) {
+        printf("[ComputeCameraCovariance] Error opening file %s for writing\n",
+               "covariance.txt");
+        return;
+    }
+
     int count = 0;
     for (int i = 0; i < num_images; i++) {
         if (m_image_data[i].m_camera.m_adjusted) {
@@ -2232,7 +2382,17 @@ void SkeletalApp::ComputeCameraCovariance()
             printf("Covariance [Image %d]:\n", i);
             matrix_print(3, 3, C);
 
+            fprintf(f, "%d\n", i);
+            fprintf(f, "%0.6f %0.6f %0.6f "
+                    "%0.6f %0.6f %0.6f "
+                    "%0.6f %0.6f %0.6f\n", 
+                    C[0], C[1], C[2], C[3], C[4], C[5], C[6], C[7], C[8]);
+            fprintf(f, "%0.6f\n", C[0] + C[4] + C[8]);
+
             count++;
         }
     }
+
+    fclose(f);
 }
+#endif
