@@ -77,6 +77,7 @@ public:
         // m_start_camera = -1;
         m_min_track_views = 2;
         m_max_track_views = 100000;
+        m_min_num_feat_matches = 16;
         m_min_max_matches = 16;
         m_ray_angle_threshold = 2.0;
 
@@ -129,9 +130,13 @@ public:
         m_scale_focal = 1.0;
         m_scale_focal_file = NULL;
         m_rotate_cameras_file = NULL;
+        m_output_relposes = false;
+        m_output_relposes_file = NULL;
 
 
 
+        m_sky_model_file = NULL;
+
         m_compute_covariance = false;
         m_covariance_fix1 = -1;
         m_covariance_fix2 = -1;
@@ -280,7 +285,13 @@ public:
 				    double max_reprojection_error = 16.0,
                                     int min_views = 2);
     
-    
+    /* Remove bad points and cameras from a reconstruction */
+    int RemoveBadPointsAndCameras(int num_points, int num_cameras, 
+                                  int *added_order, 
+                                  camera_params_t *cameras,
+                                  v3_t *points, v3_t *colors,
+                                  std::vector<ImageKeyVector> &pt_views);
+
     /* Compute pose of all cameras */
     void BundleAdjust();
 
@@ -299,7 +310,8 @@ public:
 
     /* Pick a good initial pair of cameras to bootstrap the bundle
      * adjustment */
-    void BundlePickInitialPair(int &i_best, int &j_best);
+    void BundlePickInitialPair(int &i_best, int &j_best, 
+                               bool use_init_focal_only);
 
     /* Setup the initial camera pair for bundle adjustment */
     int SetupInitialCameraPair(int i_best, int j_best,
@@ -383,7 +395,7 @@ public:
 #endif /* __DEMO__ */
 
     bool BundleTwoFrame(int i1, int i2, TwoFrameModel *model, 
-                        double &angle_out, int &num_pts_out, bool &p_edge,
+                        double &angle_out, int &num_pts_out, 
                         bool bundle_from_tracks);
     bool EstimateRelativePose(int i1, int i2, 
                               camera_params_t &camera1, 
@@ -412,10 +424,18 @@ public:
     void RotateCameras(char *rotate_file);
     void PruneBadPoints();
     void ZeroDistortionParams();
+    void OutputRelativePoses2D(const char *outfile);
+    void OutputRelativePoses3D(const char *outfile);
+    void ComputeCameraCovariance();
 
     /* Analyze point statistics */
     void AnalyzePoints();
 
+    /* Find a ground plane in the scene */
+    void FindGroundPlane();
+    /* Find a sky plane in the scene */
+    void FindSkyPlane();
+
     /* Coalesce feature descriptors for each feature point */
     void CoalesceFeatureDescriptors();
     void CoalesceFeatureDescriptorsMedian();
@@ -444,6 +464,7 @@ public:
 
 
 
+
     /* **** Bundler Options **** */
 
     bool m_panorama_mode;        /* Are we reconstructing a panorama? */
@@ -573,6 +594,12 @@ public:
     char *m_rotate_cameras_file;
     char *m_track_file;
 
+    bool m_output_relposes;
+    char *m_output_relposes_file;
+
+    bool m_segment_sky;          /* Activative sky segmentation */
+    char *m_sky_model_file; 
+
 
 
     bool m_enrich_points;        /* Enrich the point set? */
@@ -584,6 +611,7 @@ public:
 };
 
 
+typedef std::pair<int,int> IntPair;
 
 class SkeletalApp : public BundlerApp
 {
@@ -598,16 +626,38 @@ public:
     /* Process command line options */
     virtual void ProcessOptions(int argc, char **argv);
 
+    /* Estimate the global scene orientation based on geotags */
+    void EstimateGlobalOrientation(ModelMap &models);
+
+    /* Read in a set of initial rotations */
+    void ReadInitialRotations(const char *filename, double *Rout);
+    void ReadInitialTranslations(const char *filename, double *tout);
+    void ReadGlobalOrientation(const char *filename, double *R);
+    void WriteGlobalOrientation(const char *filename, double *R);
+
+    /* Refine the current set of rotations using pairwise rotations */
+    void RefineInitialRotations(ModelMap &models);
+    void RefineInitialTranslations(ModelMap &models);
+
+    /* Read in a set of rotations produced by a previous run */
+    void ReadRotationsFile(const char *filename, double *Rout);
+    
+    /* Estimate the confidences of geotags based on agreement with a
+     * global translation */
+    void ComputeGeotagConfidence(std::vector<IntPair> &pairs,
+                                 std::vector<bool> &pairs_correct,
+                                 std::vector<double> &confidence);
+
 #ifndef __DEMO__
     /* Two-frame bundle adjustment */
-    ModelMap BundleAllPairs(char *out_file, PEdgeMap &p_edges_out,
+    ModelMap BundleAllPairs(char *out_file, 
                             bool bundle_from_tracks, bool detect_duplicates);
     // ModelMap ReadModels(FILE *f);
 
     /* Estimate a similarity transform between two 2-frame models */
     bool EstimateSimilarityTransform(const TwoFrameModel &m0, 
                                      const TwoFrameModel &m1, 
-                                     int idx1, int idx2,
+                                     MatchIndex idx1, MatchIndex idx2,
                                      // double *S, 
                                      double &scale,
                                      bool verbose = false,
@@ -628,7 +678,6 @@ public:
                              int second_node = -1, int max_depth = -1,
                              double bound_factor = 1.0);
 
-    void ComputeCameraCovariance();
     void PrunePairGraph(ModelMap &models, PEdgeMap &p_edges, 
                         double prune_factor);
     void ConstructTSpanner(ModelMap &models, double t);
@@ -647,7 +696,8 @@ public:
     void NodeConnectedThrough(int image, int link, NodeState *states,
                               ModelMap &models, ModelMap &subgraph,
                               int &num_connected, int &num_total);
-    bool EdgeCanBePruned(ModelMap &models, PEdgeMap &p_edges, int i1, int i2);
+    bool EdgeCanBePruned(ModelMap &models, PEdgeMap &p_edges, 
+                         unsigned int i1, unsigned int i2);
     void ColorIncidentEdgesGray(ModelMap &models, NodeState *states, 
                                 int i1, int i2, bool symmetric = false);
     void ColorPNodes(ModelMap &models, PEdgeMap &p_edges, bool *p_node_flags, 
@@ -701,7 +751,24 @@ public:
 
     bool m_detect_duplicates;    /* ... when computing pairwise recons. */
 
-    int m_start_camera;          /* Camera to seed the t-spanner */
+    bool m_dump_pairs_sparse;    /* Write the pairs file to a nicer format */
+    bool m_estimate_orientation; /* Estimate global orientation */
+    bool m_threshold_twists;     /* Get rid of cameras with large
+                                  * twist */
+
+    char *m_initial_rotations_file;  /* Refine rotation matrices using
+                                      * this file for initialization */
+    char *m_initial_translations_file;  /* Refine translations using
+                                         * this file for initialization */
+    char *m_rotations_file;      /* Refined rotations produced on a
+                                  * previous run */
+    char *m_global_orientation_file;  /* Contains global scene
+                                       * orientation */
+
+    char *m_pairs_file;          /* File to read the pairwise
+                                  * reconstructions from */
+
+    // int m_start_camera;          /* Camera to seed the t-spanner */
 };
 
 #endif /* __bundlerapp_h__ */
